use std::fmt;

use serde::{Deserialize, Serialize};
use thiserror::Error;

// TODO: Custom serializer/deserializer for format namespace:path as string
#[derive(Debug, Clone, PartialEq, Eq, Hash, Serialize, Deserialize)]
#[serde(try_from = "&str", into = "String")]
pub struct Identifier {
    pub namespace: String,
    pub path: String,
}

impl Identifier {
    pub fn new(namespace: String, path: String) -> Self {
        // TODO: Validate namespace and path
        Self { namespace, path }
    }

    pub fn minecraft(path: String) -> Self {
        Self::new("minecraft".to_string(), path)
    }

    pub fn new_str(namespace: &str, path: &str) -> Self {
        Self::new(namespace.to_string(), path.to_string())
    }

    pub fn minecraft_str(path: &str) -> Self {
        Self::minecraft(path.to_string())
    }

    pub fn from_raw_str(raw: &str) -> Result<Self, IdentifierError> {
        let mut parts = raw.split(":");
        Ok(Self::new(
            parts
                .next()
                .ok_or(IdentifierError::InvalidNamespace)?
                .to_string(),
            parts
                .next()
                .ok_or(IdentifierError::InvalidPath)?
                .to_string(),
        ))
    }
}

impl fmt::Display for Identifier {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}:{}", self.namespace, self.path)
    }
}

impl From<Identifier> for String {
    fn from(value: Identifier) -> Self {
        (&value).into()
    }
}

impl From<&Identifier> for String {
    fn from(value: &Identifier) -> Self {
        format!("{}:{}", value.namespace, value.path)
    }
}

impl TryFrom<&str> for Identifier {
    type Error = IdentifierError;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        Identifier::from_raw_str(value)
    }
}

#[derive(Debug, Error)]
pub enum IdentifierError {
    #[error("Invalid namespace")]
    InvalidNamespace,
    #[error("Invalid path")]
    InvalidPath,
}
